@@ -20,8 +20,8 @@ def __isjson(myjson):
 
 class NcmClient:
     def __init__(self,
-                 api_keys={},
-                 logEvents=True,
+                 api_keys=None,
+                 log_events=True,
                  retries=5,
                  retry_backoff_factor=2,
                  retry_on=[
@@ -41,7 +41,9 @@ class NcmClient:
         :param retry_on: types of errors on which automatic retry will occur. Optional.
         :param base_url: # base url for calls. Configurable for testing. Optional.
         """
-        self.logEvents = logEvents
+        if api_keys is None:
+            api_keys = {}
+        self.logEvents = log_events
         self.base_url = base_url
         self.session = Session()
         self.adapter = HTTPAdapter(
@@ -416,6 +418,45 @@ class NcmClient:
         result = self.__returnhandler(ncm.status_code, ncm.json(), call_type)
         return result
 
+    def patch_configuration_managers(self, router_id, configman_json):
+        """
+        This method patches an configuration_managers for associated id.
+        :param router_id: ID of router to update
+        :param configman_json: JSON of the "configuration" field of the configuration manager
+        :return:
+        """
+        call_type = 'Configuration Manager'
+
+        response = self.session.get('{0}/configuration_managers/?router.id={1}&fields=id'.format(
+            self.base_url, str(router_id)))  # Get Configuration Managers ID for current Router from API
+        response = json.loads(response.content.decode("utf-8"))  # Decode the response and make it a dictionary
+        configman_id = response['data'][0]['id']  # get the Configuration Managers ID from response
+
+        payload = configman_json
+
+        ncm = self.session.patch('{0}/configuration_managers/{1}/'.format(self.base_url, str(configman_id)),
+                                 data=json.dumps(payload))  # Patch indie config with new values
+
+        result = self.__returnhandler(ncm.status_code, ncm.text, call_type)
+        return result
+
+    def patch_group_configuration(self, group_id, config_json):
+        """
+        This method patches an configuration_managers for associated id.
+        :param group_id: ID of group to update
+        :param config_json: JSON of the "configuration" field of the configuration manager
+        :return:
+        """
+        call_type = 'Configuration Manager'
+
+
+        payload = config_json
+
+        ncm = self.session.patch('{0}/groups/{1}/'.format(self.base_url, str(group_id)),
+                                 data=json.dumps(payload))  # Patch indie config with new values
+        result = self.__returnhandler(ncm.status_code, ncm.text, call_type)
+        return result
+
     def copy_router_configuration(self, src_router_id, dst_router_id):
         """
         Copies the Configuration Manager config of one router to another.
@@ -437,7 +478,7 @@ class NcmClient:
 
         puturl = '{0}/configuration_managers/{1}/'.format(self.base_url, dst_configman_id)
 
-        ncm = self.session.put(puturl, data=src_config)
+        ncm = self.session.patch(puturl, data=src_config)
         result = self.__returnhandler(ncm.status_code, ncm.json(), call_type)
         return result
 
@@ -683,11 +724,42 @@ class NcmClient:
         call_type = 'Historical Locations'
         geturl = '{0}/historical_locations/?router={1}'.format(self.base_url, router_id)
 
-        allowed_params = ['created_at__gt', 'created_at_timeuuid__gt', 'created_at__lte', 'order_by', 'limit', 'offset']
+        allowed_params = ['created_at__gt', 'created_at_timeuuid__gt', 'created_at__lte', 'fields', 'limit', 'offset']
         params = self.__parse_kwargs(kwargs, allowed_params)
 
         return self.__get_json(geturl, call_type, params=params)
 
+    def get_historical_locations_for_date(self, router_id, date, tzoffset_hrs=0, limit='all', **kwargs):
+        """
+        This method provides a history of device alerts. To receive device alerts, you must enable them
+        through the ECM UI: Alerts -> Settings. The info section of the alert is firmware dependent and
+        may change between firmware releases.
+        :param router_id: ID of the router
+        :param date: Date to filter logs. Must be in format "YYYY-mm-dd"
+        :type date: str
+        :param tzoffset_hrs: Offset from UTC for local timezone
+        :type tzoffset_hrs: int
+        :param limit: Number of records to return. Specifying "all" returns all records. Default all.
+        :param kwargs: A set of zero or more allowed parameters in the allowed_params list.
+        :return:
+        """
+
+        d = datetime.strptime(date, '%Y-%m-%d') + timedelta(hours=tzoffset_hrs)
+        start = d.strftime("%Y-%m-%dT%H:%M:%S")
+        end = (d + timedelta(hours=24)).strftime("%Y-%m-%dT%H:%M:%S")
+
+        call_type = 'Historical Locations'
+        geturl = '{0}/historical_locations/?router={1}'.format(self.base_url, router_id)
+
+        allowed_params = ['created_at__gt', 'created_at_timeuuid__gt', 'created_at__lte', 'fields', 'limit', 'offset']
+        params = self.__parse_kwargs(kwargs, allowed_params)
+
+        params.update({'created_at__lte': end,
+                       'created_at__gt': start,
+                       'limit': limit})
+
+        return self.__get_json(geturl, call_type, params=params)
+
     def get_locations(self, **kwargs):
         """
         This method gives a list of locations.
@@ -697,11 +769,59 @@ class NcmClient:
         call_type = 'Locations'
         geturl = '{0}/locations/'.format(self.base_url)
 
-        allowed_params = ['id', 'id__in', 'limit', 'offset']
+        allowed_params = ['id', 'id__in', 'router', 'limit', 'offset']
         params = self.__parse_kwargs(kwargs, allowed_params)
 
         return self.__get_json(geturl, call_type, params=params)
 
+    def create_location(self, account_id, latitude, longitude, router_id):
+        """
+        This method creates a location and applies it to a router.
+        :param account_id: Account which owns the object
+        :param latitude: A device's relative position north or south on the Earth's surface,
+        in degrees from the Equator
+        :param longitude: A device's relative position east or west on the Earth's surface,
+        in degrees from the prime meridian
+        :param router_id: Device that the location is associated with
+        :return:
+        """
+
+        call_type = 'Locations'
+        posturl = '{0}/locations/'.format(self.base_url)
+
+        postdata = {
+            'account': 'https://www.cradlepointecm.com/api/v2/accounts/{}/'.format(str(account_id)),
+            'accuracy': 0,
+            'latitude': latitude,
+            'longitude': longitude,
+            'method': 'manual',
+            'router': 'https://www.cradlepointecm.com/api/v2/routers/{}/'.format(str(router_id))
+        }
+
+        ncm = self.session.post(posturl, data=json.dumps(postdata))
+        result = self.__returnhandler(ncm.status_code, ncm.json(), call_type)
+        return result
+
+    def delete_location_for_router(self, router_id):
+        """
+        This operation deletes the location for a router by ID.
+        :param router_id: ID of router for which to remove location.
+        :return:
+        """
+        call_type = 'Locations'
+
+        locations = self.get_locations(router=router_id)
+        if locations:
+            location_id = locations[0]['id']
+
+            posturl = '{0}/locations/{1}/'.format(self.base_url, location_id)
+
+            ncm = self.session.delete(posturl)
+            result = self.__returnhandler(ncm.status_code, ncm.text, call_type)
+            return result
+        else:
+            return "NO LOCATION FOUND"
+
     def get_net_device_health(self, **kwargs):
         """
         This operation gets cellular heath scores, by device.
@@ -732,6 +852,34 @@ class NcmClient:
 
         return self.__get_json(geturl, call_type, params=params)
 
+    def get_net_devices_metrics_for_wan(self, **kwargs):
+        """
+        This endpoint is supplied to allow easy access to the latest signal and usage data reported by an account’s
+        net_devices without querying the historical raw sample tables, which are not optimized for a query spanning
+        many net_devices at once. Returns data only for WAN interfaces.
+        :param kwargs: A set of zero or more allowed parameters in the allowed_params list.
+        :return:
+        """
+        ids = []
+        for net_device in self.get_net_devices(mode='wan'):
+            ids.append(net_device['id'])
+        idstring = ','.join(str(x) for x in ids)
+        return self.get_net_device_metrics(net_device__in=idstring, **kwargs)
+
+    def get_net_devices_metrics_for_mdm(self, **kwargs):
+        """
+        This endpoint is supplied to allow easy access to the latest signal and usage data reported by an account’s
+        net_devices without querying the historical raw sample tables, which are not optimized for a query spanning
+        many net_devices at once. Returns data only for Modem interfaces.
+        :param kwargs: A set of zero or more allowed parameters in the allowed_params list.
+        :return:
+        """
+        ids = []
+        for net_device in self.get_net_devices(is_asset=True):
+            ids.append(net_device['id'])
+        idstring = ','.join(str(x) for x in ids)
+        return self.get_net_device_metrics(net_device__in=idstring, **kwargs)
+
     def get_net_device_signal_samples(self, **kwargs):
         """
         This endpoint is supplied to allow easy access to the latest signal and usage data reported by an account’s
@@ -778,48 +926,20 @@ class NcmClient:
         geturl = '{0}/net_devices/'.format(self.base_url)
 
         allowed_params = ['account', 'account__in', 'connection_state', 'connection_state__in', 'id', 'id__in',
-                          'is_asset', 'ipv4_address', 'ipv4_address', 'mode', 'mode__in', 'router', 'router__in',
+                          'is_asset', 'ipv4_address', 'ipv4_address__in', 'mode', 'mode__in', 'router', 'router__in',
                           'expand', 'limit', 'offset']
         params = self.__parse_kwargs(kwargs, allowed_params)
 
         return self.__get_json(geturl, call_type, params=params)
 
-    # TODO handle kwargs
     def get_net_devices_for_router(self, router_id, **kwargs):
         """
         This method gives a list of net devices for a given router.
         :param router_id: ID of the router
         :return:
         """
-        return self.get_net_devices(router=router_id)
+        return self.get_net_devices(router=router_id, **kwargs)
 
-    def get_net_devices_metrics_for_wan(self, **kwargs):
-        """
-        This endpoint is supplied to allow easy access to the latest signal and usage data reported by an account’s
-        net_devices without querying the historical raw sample tables, which are not optimized for a query spanning
-        many net_devices at once. Returns data only for WAN interfaces.
-        :param kwargs: A set of zero or more allowed parameters in the allowed_params list.
-        :return:
-        """
-        ids = []
-        for net_device in self.get_net_devices(mode='wan'):
-            ids.append(net_device['id'])
-        idstring = ','.join(str(x) for x in ids)
-        return self.get_net_device_metrics(net_device__in=idstring)
-
-    def get_net_devices_metrics_for_mdm(self, **kwargs):
-        """
-        This endpoint is supplied to allow easy access to the latest signal and usage data reported by an account’s
-        net_devices without querying the historical raw sample tables, which are not optimized for a query spanning
-        many net_devices at once. Returns data only for Modem interfaces.
-        :param kwargs: A set of zero or more allowed parameters in the allowed_params list.
-        :return:
-        """
-        ids = []
-        for net_device in self.get_net_devices(is_asset=True):
-            ids.append(net_device['id'])
-        idstring = ','.join(str(x) for x in ids)
-        return self.get_net_device_metrics(net_device__in=idstring)
 
     def get_net_devices_for_router_by_mode(self, router_id, mode, **kwargs):
         """
@@ -829,7 +949,7 @@ class NcmClient:
         :param kwargs: A set of zero or more allowed parameters in the allowed_params list.
         :return:
         """
-        return self.get_net_devices(router=router_id, mode=mode)
+        return self.get_net_devices(router=router_id, mode=mode, **kwargs)
 
     def get_products(self, **kwargs):
         """
@@ -1081,10 +1201,11 @@ class NcmClient:
         call_type = 'Routers'
         geturl = '{0}/routers/'.format(self.base_url)
 
-        allowed_params = ['account', 'account__in', 'group', 'group__in', 'id', 'id__in',
-                          'ipv4_address', 'ipv4_address__in', 'mac', 'mac__in', 'name', 'name__in', 'state',
-                          'state__in', 'state_updated_at__lt', 'state_updated_at__gt', 'updated_at__lt',
-                          'updated_at__gt', 'expand', 'order_by', 'limit', 'offset']
+        allowed_params = ['account', 'account__in', 'device_type', 'device_type__in', 'group', 'group__in', 'id',
+                          'id__in', 'ipv4_address', 'ipv4_address__in', 'mac', 'mac__in', 'name', 'name__in',
+                          'reboot_required', 'reboot_required__in', 'state', 'state__in', 'state_updated_at__lt',
+                          'state_updated_at__gt', 'updated_at__lt', 'updated_at__gt', 'expand', 'order_by', 'limit',
+                          'offset']
         params = self.__parse_kwargs(kwargs, allowed_params)
 
         return self.__get_json(geturl, call_type, params=params)
@@ -1153,6 +1274,44 @@ class NcmClient:
         return self.rename_router_by_id(
             self.get_router_by_name(existing_router_name)['id'], new_router_name)
 
+    def assign_router_to_group(self, router_id, group_id):
+        """
+        This operation assigns a router to a group.
+        :param router_id: ID of router to move.
+        :param group_id: ID of destination group.
+        :return:
+        """
+        call_type = "Routers"
+
+        puturl = '{0}/routers/{1}/'.format(self.base_url, str(router_id))
+
+        putdata = {
+            "group": 'https://www.cradlepointecm.com/api/v2/groups/{}/'.format(group_id)
+        }
+
+        ncm = self.session.put(puturl, data=json.dumps(putdata))
+        result = self.__returnhandler(ncm.status_code, ncm.json(), call_type)
+        return result
+
+    def assign_router_to_account(self, router_id, account_id):
+        """
+        This operation assigns a router to an account.
+        :param router_id: ID of router to move.
+        :param account_id: ID of destination account.
+        :return:
+        """
+        call_type = "Routers"
+
+        puturl = '{0}/routers/{1}/'.format(self.base_url, str(router_id))
+
+        putdata = {
+            "account": 'https://www.cradlepointecm.com/api/v2/accounts/{}/'.format(account_id)
+        }
+
+        ncm = self.session.put(puturl, data=json.dumps(putdata))
+        result = self.__returnhandler(ncm.status_code, ncm.json(), call_type)
+        return result
+
     def delete_router_by_id(self, router_id):
         """
         This operation deletes a router by ID.
@@ -1175,39 +1334,13 @@ class NcmClient:
         return self.delete_router_by_id(
             self.get_router_by_name(router_name)['id'])
 
-    def get_speed_test(self, speed_test_id):
-        """
-        Gets the results of a speed test job. The results are updated with the latest known state of the speed tests.
-        :param speed_test_id: ID ot Speed Test
-        :return:
-        """
-        call_type = 'Speed Test'
-        geturl = '{0}/speed_test/{1}/'.format(self.base_url, str(speed_test_id))
-
-        ncm = self.session.get(geturl)
-        result = self.__returnhandler(ncm.status_code, ncm.json()['data'], call_type)
-        return result
-
-    # TODO create speed test
-
-    def delete_speed_test(self, speed_test_id):
-        """
-        Deletes a speed test job. Deleting a job aborts it, but any test already started on a router will finish.
-        :param speed_test_id: Speed Test ID to delete
-        :return:
-        """
-        call_type = 'Speed Test'
-        posturl = '{0}/speed_test/{1}'.format(self.base_url, str(speed_test_id))
-
-        ncm = self.session.delete(posturl)
-        result = self.__returnhandler(ncm.status_code, ncm.json(), call_type)
-        return result
-
-    def set_lan_ip_address(self, router_id, lan_ip):
+    def set_lan_ip_address(self, router_id, lan_ip, netmask=None, network_id=0):
         """
         This method sets the IP Address for the Primary LAN for a given router id.
         :param router_id: ID of router to update
         :param lan_ip: LAN IP Address. (e.g. 192.168.1.1)
+        :param netmask: Subnet mask. (e.g. 255.255.255.0)
+        :param network_id: The ID of the network to update. Numbering starts from 0.
         :return:
         """
         call_type = 'LAN IP Address'
@@ -1217,19 +1350,34 @@ class NcmClient:
         response = json.loads(response.content.decode("utf-8"))  # Decode the response and make it a dictionary
         configman_id = response['data'][0]['id']  # get the Configuration Managers ID from response
 
-        payload = {
-            "configuration": [
-                {
-                    "lan": {
-                        "00000000-0d93-319d-8220-4a1fb0372b51": {
-                            "_id_": "00000000-0d93-319d-8220-4a1fb0372b51",
-                            "ip_address": lan_ip
+        if netmask:
+            payload = {
+                "configuration": [
+                    {
+                        "lan": {
+                            network_id: {
+                                "ip_address": lan_ip,
+                                "netmask": netmask
+                            }
                         }
-                    }
-                },
-                []
-            ]
-        }
+                    },
+                    []
+                ]
+            }
+
+        else:
+            payload = {
+                "configuration": [
+                    {
+                        "lan": {
+                            network_id: {
+                                "ip_address": lan_ip
+                            }
+                        }
+                    },
+                    []
+                ]
+            }
 
         ncm = self.session.patch('{0}/configuration_managers/{1}/'.format(self.base_url, str(configman_id)),
                                  data=json.dumps(payload))  # Patch indie config with new values
